/*
 * This program source code file is part of SystemK, a library in the KTag project.
 *
 * 🛡️ <https://ktag.clubk.club> 🃞
 *
 * Copyright © 2016-2025 Joseph P. Kearney and the KTag developers.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
 * details.
 *
 * There should be a copy of the GNU Affero General Public License in the LICENSE
 * file in the root of this repository.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Include Files */
#include "SystemK.h"

static const char *KLOG_TAG = "STATE_PLAYING__TAGGED_OUT";

static void Reset_Bombs()
{
    uint8_t n_bombs;
    if (SETTINGS_get_uint8_t(SYSTEMK_SETTING_N_SPECIAL_WEAPONS_ON_REENTRY, &n_bombs) == SYSTEMK_RESULT_SUCCESS)
    {
        Set_Available_Bombs(n_bombs);
    }
    else
    {
        Set_Available_Bombs(0);
        KLOG_ERROR(KLOG_TAG, "Error getting available bombs!");
    }
}

static void Playing__Tagged_Out_Entry(StateMachineContext_T * context);
static void Playing__Tagged_Out_Do(StateMachineContext_T * context);
static void Playing__Tagged_Out_Exit(StateMachineContext_T * context);

//! Activities for the **Tagged Out** substate of the **Playing** state.
const StateActivity_T STATE_PLAYING__TAGGED_OUT_Activities =
{
   .Entry = Playing__Tagged_Out_Entry,
   .Do = Playing__Tagged_Out_Do,
   .Exit = Playing__Tagged_Out_Exit
};

//! Sets up the Tagged Out substate.
/*!
 *  \param  context Context in which this substate is being run.
 */
static void Playing__Tagged_Out_Entry(StateMachineContext_T * context)
{
    if (STATE_IsPlayingSubstate(context->States.Previous_State) == false)
    {
        Playing_Entry(context);
    }
    
    LOG("Entering the Tagged Out substate of the Playing state.");
    AudioAction_T audio_action = {.ID = AUDIO_PLAY_TAGGED_OUT, .Data = (void *)0x00};
    Perform_Audio_Action(&audio_action);
    NeoPixelsAction_T neopixels_action = {.ID = NEOPIXELS_TAGGED_OUT, .Prominence = NEOPIXELS_FOREGROUND, .Data = (void *)0x00};
    xQueueSend(xQueueNeoPixels, &neopixels_action, 0);

    BLE_UpdateStatusPacket(STATE_PLAYING__TAGGED_OUT);

    Increment_Times_Tagged_Out();
}

//! Executes the Tagged Out substate.
/*!
 *  \param  context Context in which this substate is being run.
 */
static void Playing__Tagged_Out_Do(StateMachineContext_T * context)
{   
    portBASE_TYPE xStatus;
    static KEvent_T Event;
        
    xStatus = Receive_KEvent(&Event);
    
    if (xStatus == pdPASS)
    {
        switch (Event.ID)
        {         
            case KEVENT_BLE_PACKET_RECEIVED:
                if (((BLE_Packet_T *)Event.Data)->Generic.type == BLE_PACKET_TYPE_TAG)
                {
                    HandleBLETagPacket((BLE_TagPacket_T *)Event.Data);
                    
                    if (Back_In() == true)
                    {
                        Reset_Bombs();
                        Transition_For_Event(context, STATE_PLAYING__INTERACTING, &Event);
                    }
                }
                else if (((BLE_Packet_T *)Event.Data)->Generic.type == BLE_PACKET_TYPE_EVENT)
                {
                    HandleBLEEventPacket((BLE_EventPacket_T *)Event.Data, context);
                }
                else
                {
                    BLE_FreePacketBuffer(Event.Data);
                }
                break;
                
            
            case KEVENT_TRIGGER_SWITCH_RELEASED:
                {
                    uint32_t duration_of_press_in_ms = (uint32_t)Event.Data;
                    
                    // Was it a "long" press?
                    if (duration_of_press_in_ms > 5000)
                    {
                        Set_Health(Get_Max_Health());
                        Reset_Bombs();
                        Transition_For_Event(context, STATE_PLAYING__INTERACTING, &Event);
                    }
                }
                break;

            case KEVENT_GAME_OVER:
                {
                    AudioAction_T audio_action = {.ID = AUDIO_PLAY_GAME_OVER, .Play_To_Completion = true};
                    Perform_Audio_Action(&audio_action);
                    Transition_For_Event(context, STATE_WRAPPING_UP, &Event);
                }
                break;
                
            case KEVENT_PLAY_PRESSED_ON_REMOTE:
                Set_Health(Get_Max_Health());
                Reset_Bombs();
                Transition_For_Event(context, STATE_PLAYING__INTERACTING, &Event);
                break;
                
            default:
                // All other events are ignored in this state.
                ProcessUnhandledEvent(&Event);
                break;
        }
    }
}

//! Cleans up the Tagged Out substate.
/*!
 *  \param  context Context in which this substate is being run.
 */
static void Playing__Tagged_Out_Exit(StateMachineContext_T * context)
{
    NeoPixelsAction_T neopixels_action = {.ID = NEOPIXELS_ALL_OFF, .Prominence = NEOPIXELS_FOREGROUND, .Data = (void *)0x00};
    xQueueSend(xQueueNeoPixels, &neopixels_action, 0);
    AudioAction_T audio_action = {.ID = AUDIO_SILENCE, .Data = (void *)0x00};
    Perform_Audio_Action(&audio_action);
    
    if (STATE_IsPlayingSubstate(context->States.Next_State) == false)
    {
        Playing_Exit(context);
    }
}
